package com.jakubu9333.grammarManager.api;

import com.fasterxml.jackson.annotation.JsonView;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.Rule;

import com.jakubu9333.grammarManager.model.View;
import com.jakubu9333.grammarManager.service.GrammarService;
import com.jakubu9333.grammarManager.service.RuleService;
import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.CrossOrigin;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.server.ResponseStatusException;

import java.util.List;
import java.util.Set;

/**
 * @author Jakub Uhlarik
 */

@RestController
@CrossOrigin(origins = "http://localhost:3000")
@RequestMapping(path = "api/rules")
public class RuleController {

    private final RuleService ruleService;
    private final GrammarService grammarService;

    @Autowired
    public RuleController(RuleService ruleService, GrammarService grammarService) {
        this.ruleService = ruleService;
        this.grammarService = grammarService;
    }


    @JsonView(View.GrammarWithoutRules.class)
    @GetMapping("all")
    public ResponseEntity<List<Rule>> getAllRules() {
        return new ResponseEntity<>(ruleService.getAllRules(), HttpStatus.OK);
    }

    @JsonView(View.GrammarWithoutRules.class)
    @PostMapping("{grammarId}")
    public ResponseEntity<Rule> addRule(@RequestBody Rule rule, @PathVariable Long grammarId) {
        Grammar grammar = grammarService.getGrammarById(grammarId);
        Pair<Rule, Boolean> result = ruleService.processAddRule(rule, grammar);
        return new ResponseEntity<>(result.getValue0(), result.getValue1() ? HttpStatus.CREATED : HttpStatus.OK);

    }

    @JsonView(View.GrammarWithoutRules.class)
    @PostMapping("{grammarId}/batch")
    public ResponseEntity<Set<Rule>> addMultipleRules(@RequestBody Set<Rule> rules, @PathVariable Long grammarId) {
        Grammar grammar = grammarService.getGrammarById(grammarId);
        Pair<Set<Rule>, Boolean> result = ruleService.addRuleMultiple(rules, grammar);
        return new ResponseEntity<>(result.getValue0(), result.getValue1() ? HttpStatus.CREATED : HttpStatus.OK);

    }

    @JsonView(View.GrammarWithoutRules.class)
    @GetMapping("/{id}")
    public ResponseEntity<Rule> getRuleById(@PathVariable Long id) {
        return new ResponseEntity<>(ruleService.getById(id), HttpStatus.OK);
    }


    @DeleteMapping("/all")
    public ResponseEntity<?> deleteAllRules() {
        ruleService.deleteRules();
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);

    }

    @DeleteMapping("/{id}")
    public ResponseEntity<?> deleteRule(@PathVariable Long id) {
        ruleService.deleteById(id);
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    }

    @JsonView(View.GrammarWithoutRules.class)
    @GetMapping("/grammar/{grammarId}")
    public ResponseEntity<Set<Rule>> getRulesFromGrammar(@PathVariable Long grammarId) {
        return new ResponseEntity<>(grammarService.getGrammarById(grammarId).getRules(), HttpStatus.OK);
    }

    @JsonView(View.GrammarWithoutRules.class)
    @GetMapping("/grammar/{grammarId}/{ruleId}")
    public ResponseEntity<Rule> getRuleFromGrammar(@PathVariable Long grammarId, @PathVariable Long ruleId) {
        Grammar grammar = grammarService.getGrammarById(grammarId);
        Rule rule = ruleService.getById(ruleId);
        if (!grammar.getRules().contains(rule)) {
            throw new ResponseStatusException(HttpStatus.NOT_FOUND, "rule with " + ruleId + " not found in grammar ");
        }
        return new ResponseEntity<>(rule, HttpStatus.OK);
    }


    @JsonView(View.GrammarWithoutRules.class)
    @GetMapping("/left")
    public ResponseEntity<List<Rule>> getRuleByLeftSide(@RequestParam() String leftSymbol) {
        return new ResponseEntity<>(ruleService.getByLeftSide(leftSymbol), HttpStatus.OK);
    }

    @JsonView(View.GrammarWithoutRules.class)
    @PutMapping("{id}")
    public ResponseEntity<Rule> updateRuleById(@PathVariable Long id, @RequestBody Rule rule) {
        return new ResponseEntity<>(ruleService.updateRule(id, rule), HttpStatus.OK);
    }


}
