package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.FileHolderRepo;
import com.jakubu9333.grammarManager.dao.GrammarRepository;
import com.jakubu9333.grammarManager.model.FileHolder;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RightSidePart;
import com.jakubu9333.grammarManager.model.Rule;
import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.transaction.Transactional;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.StringReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * @author Jakub Uhlarik
 */
@Service
public class FileService {

    private final FileHolderRepo fileHolderRepo;

    private final GrammarService grammarService;
    private final GrammarRepository grammarRepository;

    @Autowired
    public FileService(FileHolderRepo fileHolderRepo, GrammarService grammarService, GrammarRepository grammarRepository) {
        this.fileHolderRepo = fileHolderRepo;
        this.grammarService = grammarService;
        this.grammarRepository = grammarRepository;
    }

    public List<FileHolder> getFilesFromGrammar(Long id) {
        return fileHolderRepo.findAllByGrammarId(id);
    }

    public static class XmlGrammarHandler extends DefaultHandler {
        private Grammar newGrammar;
        private StringBuilder rightSidePart;

        private Rule rule = null;
        private boolean item = false;

        @Override
        public void startDocument() {
            newGrammar = new Grammar(null, new HashSet<>(), null);
        }

        @Override
        public void characters(char[] ch, int start, int length) {
            if (rightSidePart != null && item) {
                rightSidePart.append(ch, start, length);
            }
        }

        @Override
        public void startElement(String uri, String lName, String qName, Attributes attr) {
            switch (qName) {
                case "grammar" -> {
                    String grammarUri = attr.getValue("xml:base");
                    if (grammarUri != null) {
                        newGrammar.setUri(grammarUri);
                    }
                    String rootLeft = attr.getValue("root");
                    if (rootLeft != null) {
                        newGrammar.getRules().add(new Rule(newGrammar, rootLeft, new HashSet<>(), true));
                    }
                }
                case "rule" -> {
                    rule = new Rule(newGrammar, attr.getValue("id"), new HashSet<>(), false);
                    rightSidePart = new StringBuilder();
                }
                case "item" -> item = true;

                case "ruleref" -> {

                    String spRule = attr.getValue("special");
                    if (spRule != null) {
                        rightSidePart.append("<").append(spRule).append(">");
                    }
                    String ruleRef = attr.getValue("uri");
                    if (ruleRef != null) {
                        if (ruleRef.startsWith("#")) {
                            ruleRef = ruleRef.substring(1);
                        }
                        rightSidePart.append("<").append(ruleRef).append(">");
                    }

                }
            }
        }

        @Override
        public void endElement(String uri, String lName, String qName) {
            switch (qName) {
                case "rule" -> {

                    newGrammar.getRules().add(rule);
                    rule = null;
                }
                case "item" -> {
                    item = false;
                    rule.getRightSide().add(new RightSidePart(rightSidePart.toString()));
                    rightSidePart = new StringBuilder();
                }
            }
        }
    }


    @Transactional
    public void deleteFilesFromGrammar(Long grammarId) {
        Grammar grammar = grammarService.getGrammarById(grammarId);
        Set<FileHolder> files = grammar.getFiles();
        grammar.setFiles(new HashSet<>());
        fileHolderRepo.deleteAll(files);

    }

    @Transactional
    public FileHolder getGeneratedFile(Long grammarId, boolean xml) throws IOException {
        FileHolder fileHolder = fileHolderRepo.getFileHolderByGeneratedTrueAndGrammar_IdAndXml(grammarId, xml);
        if (fileHolder == null) {
            newFileFromGrammar(grammarId, xml);
            fileHolder = fileHolderRepo.getFileHolderByGeneratedTrueAndGrammar_IdAndXml(grammarId, xml);
        }
        return fileHolder;
    }

    @Transactional
    public void deleteFileById(Long id) {
        FileHolder file = fileHolderRepo.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "file:" + id + "not found"));
        file.getGrammar().getFiles().remove(file);
        fileHolderRepo.delete(file);
    }

    public List<FileHolder> getAllFiles() {
        return fileHolderRepo.findAll();
    }


    public FileHolder getFileHolderById(Long id) {
        return fileHolderRepo.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "file id:" + id + " not found"));
    }

    @Transactional
    public Pair<Grammar, Boolean> addGrammarFromFile(String file, boolean xml, Long id, boolean generateNewOnConflict, boolean saveFile) throws IOException {
        Grammar grammar;


        if (xml) {
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();

                XmlGrammarHandler handler = new XmlGrammarHandler();
                InputSource is = new InputSource(new StringReader(file));
                saxParser.parse(is, handler);

                grammar = handler.newGrammar;

            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "parsing error", e);
            }

        } else {
            grammar = jsgfParse(file);
        }
        grammar = grammarService.preProcessGrammar(grammar);
        FileHolder fileHolder = new FileHolder(null, xml, false, null);
        String filePath = "";
        if (saveFile) {
            fileHolderRepo.save(fileHolder);
            String filePathEnd = "file" + fileHolder.getId() + (xml ? ".xml" : ".jsgf");
            filePath = "./grammars/" + filePathEnd;
            fileHolder.setFilePath(filePathEnd);
        }

        if (id == null || !grammarRepository.existsById(id)) {
            if (saveFile) {
                fileHolder.setSize(printFile(filePath, file));
                grammar.getFiles().add(fileHolder);
                fileHolder.setGrammar(grammar);
            }
            return new Pair<>(grammarService.addGrammar(grammar), true);

        }
        Grammar oldGrammar = grammarService.getGrammarById(id);
        if (oldGrammar.equals(grammar)) {
            if (saveFile) {
                fileHolder.setSize(printFile(filePath, file));
                oldGrammar.getFiles().add(fileHolder);
                fileHolder.setGrammar(oldGrammar);
            }
            return new Pair<>(oldGrammar, false);
        }
        if (generateNewOnConflict) {
            if (saveFile) {
                fileHolder.setSize(printFile(filePath, file));
                grammar.getFiles().add(fileHolder);
                fileHolder.setGrammar(grammar);
            }
            return new Pair<>(grammarService.addGrammar(grammar), true);
        }
        if (saveFile) {
            fileHolderRepo.delete(fileHolder);
        }

        return null;
    }

    private long printFile(String filePath, String content) throws IOException {
        File file = new File(filePath);

        boolean exists = file.exists();
        boolean created = file.createNewFile();
        if (!exists && !created) {
            throw new RuntimeException("file error");
        }
        try (PrintStream ps = new PrintStream(new FileOutputStream(file))) {
            ps.print(content);
        }
        return Files.size(Path.of(filePath));
    }

    private Grammar jsgfParse(String file) {
        //todo jsgf parse + kleene star + angle brackets
        Grammar grammar = new Grammar();
        boolean comment = false;
        boolean complexCommentNow;
        boolean start = false;
        String[] lines = file.split("[;\n]");
        String uncommentedLine = null;
        int i = 1;
        boolean root = false;
        boolean setRoot = false;
        for (String line : lines) {
            complexCommentNow = false;
            if (setRoot) {
                root = true;
                setRoot = false;
            }
            line = line.trim();

            if (line.contains("/*")) {
                String[] splitByComment = line.split("/\\*");
                uncommentedLine = splitByComment[0];
                comment = true;
                complexCommentNow = true;
            }
            if (line.contains("*/")) {
                if (comment) {
                    comment = false;
                } else {
                    throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "comment without start; line= " + i);
                }
                String[] splitByComment = line.split("\\*/");
                line = splitByComment[1];
                if (complexCommentNow) {
                    line = uncommentedLine + line;
                }


            }
            if (line.contains("//")) {
                String[] splitByComment = line.split("//");
                line = splitByComment[1];
                if (line.equals("root")) {
                    setRoot = true;
                }
            }
            if (line.matches("#JSGF V[0-9].[0-9]+")) {
                start = true;
                i++;
                continue;
            }
            if (line.contains("grammar ")) {
                if (grammar.getUri() != null) {
                    throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "duplicate grammar name declaration; line=" + i);
                }
                String[] splitLine = line.split(" ");
                grammar.setUri(splitLine[1]);
            }
            if (line.contains("=")) {
                //rule
                String[] splitLine = line.split("=");
                String leftSymbol = splitLine[0].trim();
                RightSidePart rightSidePart = new RightSidePart(splitLine[1].trim());
                Rule rule = new Rule(null, grammar, leftSymbol, Set.of(rightSidePart), root);
                root = false;
                if (grammar.getRules() == null) {
                    grammar.setRules(new HashSet<>());
                }
                grammar.getRules().add(rule);
            }
            i++;
        }
        return grammar;
    }

    private void printRightSide(PrintStream bf, String rightSide) {
        if (rightSide.contains("<") && rightSide.contains(">")) {

            int leftindex = 0;
            int rightindex = 0;

            int begin = 0;
            while (leftindex >= 0 && rightindex >= 0) {
                leftindex = rightSide.indexOf('<', rightindex);
                rightindex = rightSide.indexOf('>', leftindex);
                //between two rules or begin
                if (leftindex - begin > 0) {

                    bf.print(rightSide.substring(begin, leftindex));
                }
                //after rules
                if (leftindex == -1 && rightSide.length() > begin) {
                    bf.print(rightSide.substring(begin));
                }
                if (leftindex >= 0 && rightindex >= 0) {
                    String ruleInside = rightSide.substring(leftindex + 1, rightindex);
                    if (ruleInside.equals("NULL")) {
                        bf.print("<ruleref special=\"NULL\"/>");
                    } else if (ruleInside.equals("VOID")) {
                        bf.print("<ruleref special=\"VOID\"/>");
                    } else {
                        bf.print("<ruleref uri=\"#" + ruleInside + "\"/>");
                    }

                }
                begin = rightindex + 1;

            }
            return;
        }
        if (!rightSide.contains("<") && !rightSide.contains(">")) {
            bf.println(rightSide);
            return;
        }
        throw new ResponseStatusException(HttpStatus.EXPECTATION_FAILED, "bad rule format");
    }

    private void writeRule(PrintStream bf, Rule rule, boolean xml) {
        if (!xml) {
            if (rule.getRoot()) {
                bf.println("//root;");
            }
            int i = 0;
            StringBuilder rightSide = new StringBuilder();
            Set<RightSidePart> setRightSide = rule.getRightSide();
            for (RightSidePart part : setRightSide) {
                rightSide.append(part.getPart());
                if (i + 1 < setRightSide.size()) {
                    rightSide.append(" | ");
                }
                i++;
            }
            bf.println(rule.getLeftSymbol() + " = " + rightSide + ";");
            return;
        }
        bf.print("<rule id=\"");
        bf.println(nameFromLeftSymbol(rule) + "\">");
        Set<RightSidePart> rightSide = rule.getRightSide();
        if (rightSide.size() > 1) {
            bf.println("<one-of>");
            for (RightSidePart part : rightSide) {
                bf.print("<item>");
                printRightSide(bf, part.getPart());
                bf.println("</item>");
            }
            bf.println("</one-of>");
        } else {
            bf.print("<item>");
            printRightSide(bf, rightSide.stream().map(RightSidePart::getPart).toList().get(0));
            bf.println("</item>");
        }
        bf.println("</rule>");
    }

    private String nameFromLeftSymbol(Rule rule) {
        String oldLeftSymbol = rule.getLeftSymbol();
        return rule.getLeftSymbol().substring(1, oldLeftSymbol.length() - 1);
    }

    private void writeHeader(PrintStream ps, Grammar grammar, boolean xml) {
        //jsgf
        if (!xml) {
            ps.println("#JSGF v 1.0;");
            ps.println("grammar " + grammar.getUri() + ";\n");
            return;
        }
        //xml
        ps.println("<?xml version=\"1.0\"?>");
        ps.println("<grammar version=\"1.0\"");
        Rule rootRule;
        String uri;
        if ((uri = grammar.getUri()) != null) {
            ps.println("xml:base=\"" + uri + "\"");
        }
        if ((rootRule = grammar.getRootRule()) != null) {
            ps.println("root=\"" + nameFromLeftSymbol(rootRule) + "\"");
        }

        ps.println(" xmlns=\"http://www.w3.org/2001/06/grammar\">");
    }

    private void generateFile(String filePath, Grammar grammar, boolean xml) throws IOException {
        File file = new File(filePath);

        boolean exists = file.exists();
        boolean created = file.createNewFile();
        if (!exists && !created) {
            throw new RuntimeException("file error");
        }
        try (PrintStream ps = new PrintStream(new FileOutputStream(file))) {
            writeHeader(ps, grammar, xml);

            Rule rootRule;
            if (null != (rootRule = grammar.getRootRule())) {
                writeRule(ps, rootRule, xml);
            }

            for (Rule rule : grammar.getRules()) {
                if (!rule.getRoot()) {
                    writeRule(ps, rule, xml);
                }
            }
            if (xml) {
                ps.println("</grammar>");
            }
        }
    }

    @Transactional
    public Pair<FileHolder, Boolean> newFileFromGrammar(Long id, boolean xml) throws IOException {
        Grammar grammar = grammarService.getGrammarById(id);
        FileHolder fileHolder = fileHolderRepo.getFileHolderByGeneratedTrueAndGrammar_IdAndXml(id, xml);
        if (fileHolder != null && (grammar.getUpdateTime() == null || fileHolder.getCreateTime().compareTo(grammar.getUpdateTime()) >= 0)) {
            return new Pair<>(fileHolder, false);
        }
        if (fileHolder != null) {
            grammar.getFiles().remove(fileHolder);
            fileHolderRepo.delete(fileHolder);
        }

        fileHolder = new FileHolder(null, xml, true, grammar);
        fileHolderRepo.save(fileHolder);
        grammar.getFiles().add(fileHolder);
        String filePathEnd = "file" + fileHolder.getId() + (xml ? ".xml" : ".jsgf");
        fileHolder.setFilePath(filePathEnd);
        String filePath = "./grammars/" + filePathEnd;
        generateFile(filePath, grammar, xml);
        fileHolder.setSize(Files.size(Path.of(filePath)));
        return new Pair<>(fileHolder, true);
    }


}

