package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.GrammarRepository;
import com.jakubu9333.grammarManager.dao.RightSidePartRepo;
import com.jakubu9333.grammarManager.dao.RuleRepository;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RightSidePart;
import com.jakubu9333.grammarManager.model.Rule;
import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;

import javax.transaction.Transactional;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * @author Jakub Uhlarik
 */
@Service
public class RuleService {

    private final RuleRepository ruleRepository;
    private final GrammarRepository grammarRepository;
    private final RightSidePartRepo rightSidePartRepo;

    @Autowired
    public RuleService(RuleRepository ruleRepository, GrammarRepository grammarRepository, RightSidePartRepo rightSidePartRepo) {
        this.ruleRepository = ruleRepository;
        this.grammarRepository = grammarRepository;
        this.rightSidePartRepo = rightSidePartRepo;

    }

    public List<Rule> getAllRules() {
        return ruleRepository.findAll();
    }

    private Rule saveRule(Rule rule) {
        return ruleRepository.save(rule);
    }

    public String stringEncapsulation(String string) {
        if (string == null) {
            return null;
        }
        StringBuilder stringBuilder = new StringBuilder();
        if (!string.startsWith("<")) {
            stringBuilder.append('<');
        }
        stringBuilder.append(string);
        if (!string.endsWith(">")) {
            stringBuilder.append('>');
        }
        return stringBuilder.toString();

    }

    private String[] splitRightSide(String string) {
        if (string == null) {
            return new String[]{""};
        }
        if (string.contains("|")) {
            return string.split("\\|");
        }
        return new String[]{string};
    }

    private String normalBrackets(String string) {
        StringBuilder sb = new StringBuilder(string);

        int leftIndex;

        int rightIndex = -1;
        while ((leftIndex = sb.indexOf("(")) != -1) {
            int brackets1 = 0;
            for (int i = leftIndex; i < sb.length(); i++) {
                if (sb.charAt(i) == '(') {
                    brackets1++;
                }
                if (sb.charAt(i) == ')') {
                    brackets1--;
                    if (brackets1 == 0) {
                        rightIndex = i;
                        break;
                    }
                }

            }
            if (rightIndex == -1) {
                break;
            }

            String inside = sb.substring(leftIndex + 1, rightIndex);
            int outsideLeft = 0;
            int outSideRight = 0;
            for (int i = leftIndex; i >= 0; i--) {
                outsideLeft = i;
                if (sb.charAt(i) == '|') {
                    break;
                }
            }
            int brackets = 0;
            for (int i = rightIndex; i < sb.length(); i++) {
                outSideRight = i;

                if (sb.charAt(i) == '(') {
                    brackets++;
                }
                if (sb.charAt(i) == ')' && i != rightIndex) {
                    brackets--;
                }
                if (sb.charAt(i) == '|' && brackets == 0) {
                    break;
                }
            }
            if (outsideLeft != 0 && leftIndex != 0) {
                outsideLeft++;
            }

            String leftOutside = sb.substring(outsideLeft, leftIndex);
            String rightOutside = sb.substring(rightIndex + 1, outSideRight + 1);
            sb.replace(outsideLeft, outSideRight + 1, "|");
            for (String stringInSplit : splitRightSide(inside)) {
                sb.append("|").append(leftOutside).append(stringInSplit).append(rightOutside);
            }


        }
        return sb.toString();
    }

    private String straightBrackets(String string) {
        StringBuilder sb = new StringBuilder(string);

        int leftIndex;

        int rightIndex;
        while ((leftIndex = sb.indexOf("[")) != -1 && (rightIndex = sb.indexOf("]")) != -1) {

            String inside = sb.substring(leftIndex + 1, rightIndex);
            sb.replace(leftIndex, rightIndex + 1, "(" + inside + "|<NULL>)");

        }
        return sb.toString();

    }

    private List<Pair<Integer, Integer>> getIgnoreListCheckRuleRef(String addedRightSide) {
        List<Pair<Integer, Integer>> result = new ArrayList<>();
        Integer start = null;
        Integer end = null;
        Integer startAngleBracket = null;
        Integer endAngleBracket = null;
        for (int i = 0; i < addedRightSide.length(); i++) {
            boolean slashNotBefore = ((i == 0 || addedRightSide.charAt(i - 1) != '\\') || (i > 1 && addedRightSide.charAt(i - 1) == '\\' && addedRightSide.charAt(i - 2) == '\\'));
            if (addedRightSide.charAt(i) == '"' && slashNotBefore) {
                if (start == null) {
                    end = null;
                    start = i;
                } else {
                    result.add(new Pair<>(start, i));
                    start = null;
                    end = i;
                }
            }
            if (addedRightSide.charAt(i) == '<' && slashNotBefore) {
                endAngleBracket = null;
                startAngleBracket = i;
            }
            if (addedRightSide.charAt(i) == '>' && slashNotBefore) {
                if (startAngleBracket == null) {
                    throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "ending angle bracket without starting bracket");
                }
                endAngleBracket = i;
                result.add(new Pair<>(startAngleBracket, endAngleBracket));
                startAngleBracket = null;

            }
        }
        if (end == null && start != null) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "last quote not ended");
        }
        if (endAngleBracket == null && startAngleBracket != null) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "ruleref not ended");
        }
        return result;
    }


    public Set<RightSidePart> stringToRightPart(String string) {
        String addedRightSide = string == null ? "<NULL>" : string;
        List<Pair<Integer, Integer>> ignoreList = getIgnoreListCheckRuleRef(addedRightSide);
        boolean straightBrackets = (addedRightSide.contains("[")) && (addedRightSide.contains("]"));
        if (straightBrackets) {
            addedRightSide = straightBrackets(addedRightSide);
        }
        boolean normalBrackets = (addedRightSide.contains("(")) && (addedRightSide.contains(")"));
        if (normalBrackets) {
            addedRightSide = normalBrackets(addedRightSide);
        }
        return Arrays.stream(addedRightSide.split("\\|")).map(String::trim).filter(p -> p.length() > 0).map(RightSidePart::new).collect(Collectors.toSet());
    }


    public Set<Rule> commonLeftMerge(Set<Rule> rules) {
        Map<String, Rule> calcMap = new HashMap<>();
        for (Rule rule : rules) {
            Rule oldValue = calcMap.get(rule.getLeftSymbol());
            if (oldValue == null) {
                calcMap.put(rule.getLeftSymbol(), rule);
            } else {
                Set<RightSidePart> leftSide = new HashSet<>();
                leftSide.addAll(oldValue.getRightSide());
                leftSide.addAll(rule.getRightSide());
                Rule mergeRule = new Rule(null, rule.getLeftSymbol(), leftSide, rule.getRoot() || oldValue.getRoot());
                calcMap.replace(rule.getLeftSymbol(), mergeRule);
            }
        }
        if (calcMap.size() == rules.size()) {
            return rules;
        }
        return new HashSet<>(calcMap.values());
    }

    public void rulePreProcess(Rule rule, boolean ignoreLeft) {
        if (ignoreLeft) {
            Set<RightSidePart> newRightSide = new HashSet<>();
            for (RightSidePart rightSidePart : rule.getRightSide()) {
                newRightSide.addAll(stringToRightPart(rightSidePart.getPart()));
            }
            rule.setRightSide(newRightSide);
            return;
        }
        rulePreProcess(rule);

    }
    
    public void rulePreProcess(Rule rule) {
        String leftSymbol = rule.getLeftSymbol();
        if (leftSymbol == null) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "no left symbol");
        }
        if (leftSymbol.trim().length() == 0) {
            rule.setLeftSymbol("empty");
        }
        rule.setLeftSymbol(stringEncapsulation(rule.getLeftSymbol().trim()));
        Set<RightSidePart> newRightSide = new HashSet<>();
        for (RightSidePart rightSidePart : rule.getRightSide()) {
            newRightSide.addAll(stringToRightPart(rightSidePart.getPart()));
        }
        rule.setRightSide(newRightSide);

    }

    private Set<RightSidePart> newRightSide(Rule oldRule, Rule newRule) {
        Set<RightSidePart> oldRightSide = oldRule.getRightSide() == null ? new HashSet<>() : oldRule.getRightSide();
        Set<RightSidePart> newRightSide = newRule.getRightSide() == null ? new HashSet<>() : newRule.getRightSide();
        newRightSide.removeAll(oldRightSide);
        newRule.setRightSide(null);
        rightSidePartRepo.deleteAll(newRightSide.stream().filter(p -> p.getId() != null).collect(Collectors.toSet()));
        return newRightSide.stream().map(p -> p.getId() != null ? new RightSidePart(p.getPart()) : p).collect(Collectors.toSet());
    }

    @Transactional
    public Pair<Rule, Boolean> findOlderRule(Rule rule, Grammar grammar) {

        boolean isRoot = rule.getRoot();

        //checks if any in this grammar rule have same left side
        List<Rule> oldRules = getByLeftSide(rule.getLeftSymbol(), grammar.getId());
        if (!oldRules.isEmpty() && !Objects.equals(oldRules.get(0).getId(), rule.getId())) {
            Rule oldRule = oldRules.get(0);
            Set<RightSidePart> newRightSide = newRightSide(oldRule, rule);
            rightSidePartRepo.saveAll(newRightSide);
            oldRule.getRightSide().addAll(newRightSide);
            if (grammar.getRootRule() == null && isRoot) {
                oldRule.setRoot(true);
                grammar.setRootRule(oldRule);
            }
            return new Pair<>(oldRule, false);
        }

        if (rule.getRightSide().size() == 0) {
            rule.getRightSide().add(new RightSidePart("<NULL>"));
        }
        return new Pair<>(rule, true);
    }

    @Transactional
    public Pair<Rule, Boolean> processAddRule(Rule rule, Grammar grammar) {
        rulePreProcess(rule);
        return addRule(rule, grammar);
    }

    @Transactional
    public Pair<Rule, Boolean> addRule(Rule rule, Grammar grammar) {
        Pair<Rule, Boolean> result = findOlderRule(rule, grammar);
        rule = result.getValue0();
        //if no new rule will be created
        if (!result.getValue1()) {
            return result;
        }

        if (grammar.getRootRule() != null && rule.getRoot()) {
            rule.setRoot(false);
        }
        rightSidePartRepo.saveAll(rule.getRightSide());
        rule = saveRule(rule);
        if (rule.getRoot()) {
            grammar.setRootRule(rule);
        }
        grammar.getRules().add(rule);
        rule.setGrammar(grammar);
        return new Pair<>(rule, true);
    }

    @Transactional
    public void deleteRules() {
        grammarRepository.findAll().forEach(((grammar) -> {
            grammar.setRules(null);
            grammar.setRootRule(null);
        }));
        ruleRepository.deleteAll();
        rightSidePartRepo.deleteAll();
    }

    @Transactional
    public void deleteById(Long id) {
        Rule rule = ruleRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "rule id: " + id + " not found"));
        Grammar rulesGrammar = rule.getGrammar();
        rulesGrammar.getRules().remove(rule);
        if (rule.equals(rulesGrammar.getRootRule())) {
            rulesGrammar.setRootRule(null);
        }
        ruleRepository.delete(rule);
        rightSidePartRepo.deleteAll(rule.getRightSide());
    }


    public List<Rule> getByLeftSide(String leftSymbol) {
        leftSymbol = stringEncapsulation(leftSymbol);
        return ruleRepository.findByLeftSymbol(leftSymbol);
    }

    public List<Rule> getByLeftSide(String leftSymbol, Long grammarId) {
        leftSymbol = stringEncapsulation(leftSymbol);
        return ruleRepository.findByLeftSymbolAndGrammar_Id(leftSymbol, grammarId);
    }

    public Rule getById(Long id) {
        return ruleRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND));
    }

    @Transactional
    public Rule updateRule(Long id, Rule rule) {
        Rule oldRule = this.getById(id);
        if (rule.getRoot() != null && oldRule.getRoot() != rule.getRoot()) {
            oldRule = changeRootStatus(id);
        }
        boolean changes = false;
        if (rule.getLeftSymbol() != null && rule.getLeftSymbol().length() != 0) {
            rulePreProcess(rule);
            oldRule.setLeftSymbol(rule.getLeftSymbol());
            changes = true;
        }
        if (rule.getRightSide() != null && rule.getRightSide().size() != 0) {
            if (!changes) {
                rulePreProcess(rule, true);
            }
            rightSidePartRepo.deleteAll(oldRule.getRightSide());
            rightSidePartRepo.saveAll(rule.getRightSide());
            oldRule.setRightSide(rule.getRightSide());
            changes = true;
        }
        if (changes) {
            return this.findOlderRule(oldRule, oldRule.getGrammar()).getValue0();
        }
        return oldRule;
    }

    @Transactional
    public Rule changeRootStatus(Long id) {
        Rule rule = this.getById(id);
        Grammar grammar = rule.getGrammar();
        if (rule.getRoot()) {
            grammar.setRootRule(null);
            rule.setRoot(false);
            return rule;
        }
        if (grammar.getRootRule() != null) {
            //return;
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "grammar have root terminal");
        }
        grammar.setRootRule(rule);
        rule.setRoot(true);
        return rule;

    }

    public Pair<Set<Rule>, Boolean> addRuleMultiple(Set<Rule> rules, Grammar grammar) {
        for (Rule rule : rules) {
            rulePreProcess(rule);
        }
        rules = commonLeftMerge(rules);

        Boolean change = false;
        Set<Rule> result = new HashSet<>();
        for (Rule rule : rules) {
            Pair<Rule, Boolean> partResult = addRule(rule, grammar);
            change |= partResult.getValue1();
            result.add(partResult.getValue0());
        }
        return new Pair<>(result, change);

    }
}
