package com.jakubu9333.grammarManager;

import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RightSidePart;
import com.jakubu9333.grammarManager.model.Rule;
import com.jakubu9333.grammarManager.service.GrammarService;
import com.jakubu9333.grammarManager.service.RuleService;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.server.ResponseStatusException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * @author Jakub Uhlarik
 */
@SpringBootTest
public class RuleServiceTest {

    @Autowired
    private RuleService ruleService;
    @Autowired
    private GrammarService grammarService;

    @Test
    public void basicRulesAdd() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "h", setOfRightSidePart("xd"), false);
        Rule rule1 = new Rule(null, "g", setOfRightSidePart("xd"), true);
        rule = ruleService.processAddRule(rule, grammar).getValue0();
        ruleService.processAddRule(rule1, grammar);
        assert grammar.getRootRule().getLeftSymbol().equals("<g>");
        assert rule.getId() != null;
        Set<Rule> rules = grammar.getRules();
        assert rules.size() == 2;
        int one = 0;
        int two = 0;
        int three = 0;
        for (Rule smallRule : rules) {
            switch (smallRule.getLeftSymbol()) {
                case "<h>" -> one++;
                case "<g>" -> two++;
                default -> three++;
            }
        }
        assert one == 1 && two == 1 && three == 0;


    }

    public static Set<RightSidePart> setOfRightSidePart(String string) {
        return new HashSet<>(List.of(new RightSidePart(string)));
    }

    @Test
    @Transactional
    public void sameLeftSideRule() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "g", setOfRightSidePart("xd"), false);
        Rule rule1 = new Rule(null, "<g>", setOfRightSidePart("xd|good"), true);
        Rule rule2 = new Rule(null, "g>", setOfRightSidePart("xd|bad"), true);
        Rule rule3 = new Rule(null, "d", setOfRightSidePart("xd|smile"), false);
        ruleService.processAddRule(rule, grammar);
        ruleService.processAddRule(rule1, grammar);
        rule = ruleService.processAddRule(rule2, grammar).getValue0();
        ruleService.processAddRule(rule3, grammar);
        int xd = 0;
        int good = 0;
        int smile = 0;
        int bad = 0;
        int def = 0;
        assert grammar.getRules().size() == 2;
        for (RightSidePart leftPart : rule.getRightSide()) {
            switch (leftPart.getPart()) {
                case "xd" -> xd++;
                case "good" -> good++;
                case "smile" -> smile++;
                case "bad" -> bad++;
                default -> def++;
            }
        }
        assert xd == 1 && good == 1 && smile == 0 && bad == 1 && def == 0;
    }

    @Test
    public void testMultiplePartSet() {
        grammarService.deleteAll();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Set<RightSidePart> input1 = new HashSet<>(List.of(new RightSidePart("xd"), new RightSidePart("xd")));
        Rule rule = new Rule(null, "b", input1, false);
        rule = ruleService.processAddRule(rule, grammar).getValue0();
        assert rule.getRightSide().equals(new HashSet<>(List.of(new RightSidePart("xd"))));

        Set<RightSidePart> input2 = new HashSet<>(List.of(new RightSidePart("xd"), new RightSidePart("dx")));
        Rule rule2 = new Rule(null, "g", input2, false);
        rule2 = ruleService.processAddRule(rule2, grammar).getValue0();
        assert rule2.getRightSide().equals(new HashSet<>(List.of(new RightSidePart("xd"), new RightSidePart("dx"))));

        Set<RightSidePart> input3 = new HashSet<>(List.of(new RightSidePart("(1|2)1"), new RightSidePart(" 21 ")));
        Rule rule3 = new Rule(null, "d", input3, false);
        rule3 = ruleService.processAddRule(rule3, grammar).getValue0();
        assert rule3.getRightSide().equals(new HashSet<>(List.of(new RightSidePart("11"), new RightSidePart("21"))));
    }

    @Test
    public void testEmptyAdd() {
        grammarService.deleteAll();
        Grammar gr = new Grammar();
        grammarService.addGrammar(gr);
        Rule nullRule = new Rule();
        Rule nullRightStr = new Rule(null, "dd", new HashSet<>(), null);
        Rule emptyStrRule = new Rule(null, "", setOfRightSidePart(""), null);
        try {
            ruleService.rulePreProcess(nullRule);
            assert false;
        } catch (ResponseStatusException ignored) {

        }
        emptyStrRule = ruleService.processAddRule(emptyStrRule, gr).getValue0();
        nullRightStr = ruleService.processAddRule(nullRightStr, gr).getValue0();
        assert emptyStrRule.getLeftSymbol().equals("<empty>");
        assert emptyStrRule.getRightSide().equals(getFromString("<NULL>"));
        assert nullRightStr.getRightSide().equals(getFromString("<NULL>"));
    }

    public static Set<RightSidePart> getFromString(String string) {
        return Arrays.stream(string.split("\\|")).map(RightSidePart::new).collect(Collectors.toSet());
    }

    @Test
    public void testNormalBrackets() {
        grammarService.deleteAll();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "g", setOfRightSidePart("(xd|dd)"), false);
        Rule rule1 = new Rule(null, "l", setOfRightSidePart("smile (good|cheer)"), false);
        Rule rule2 = new Rule(null, "p", setOfRightSidePart("(pain|bad) cry"), false);


        rule = ruleService.processAddRule(rule, grammar).getValue0();
        assert rule.getRightSide().equals(getFromString("xd|dd"));
        rule1 = ruleService.processAddRule(rule1, grammar).getValue0();
        assert rule1.getRightSide().equals(getFromString("smile good|smile cheer"));
        rule2 = ruleService.processAddRule(rule2, grammar).getValue0();
        assert rule2.getRightSide().equals(getFromString("pain cry|bad cry"));
        Rule rule3 = new Rule(null, "s", setOfRightSidePart("(a|b) (c|d) (e|f)"), false);
        rule3 = ruleService.processAddRule(rule3, grammar).getValue0();
        Set<RightSidePart> rule3RightSide = rule3.getRightSide();
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        int five = 0;
        int six = 0;
        int seven = 0;
        int eight = 0;
        int nine = 0;
        for (RightSidePart partString : rule3RightSide) {
            switch (partString.getPart()) {
                case "a c e" -> one++;
                case "a c f" -> two++;
                case "a d e" -> three++;
                case "a d f" -> four++;
                case "b c e" -> five++;
                case "b c f" -> six++;
                case "b d e" -> seven++;
                case "b d f" -> eight++;
                default -> nine++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 1 && five == 1 && six == 1 && seven == 1 && eight == 1 && nine == 0;
        Rule rule5 = new Rule(null, "5", setOfRightSidePart("sk(pain|bad)cz"), false);
        rule5 = ruleService.processAddRule(rule5, grammar).getValue0();
        assert rule5.getRightSide().equals(getFromString("skpaincz|skbadcz"));
        Rule rule6 = new Rule(null, "6", setOfRightSidePart("((1|2)|(3|4))"), false);
        rule6 = ruleService.processAddRule(rule6, grammar).getValue0();
        one = 0;
        two = 0;
        three = 0;
        four = 0;
        five = 0;
        for (RightSidePart partString : rule6.getRightSide()) {
            switch (partString.getPart()) {
                case "1" -> one++;
                case "2" -> two++;
                case "3" -> three++;
                case "4" -> four++;
                default -> five++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 1 && five == 0;

    }


    @Test
    public void testQuoteAndAngleBracket() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "0", setOfRightSidePart("\" xd \""), false);
        rule = ruleService.processAddRule(rule, grammar).getValue0();
        assert rule.getRightSide().equals(getFromString("\" xd \""));
        List<String> rightSides = new ArrayList<>();
        List<String> messages = new ArrayList<>();
        rightSides.add("\" xd \"\"");
        messages.add("last quote not ended");
        rightSides.add("<xd><");
        messages.add("ruleref not ended");
        rightSides.add("<xd><dcdc>>");
        messages.add("ending angle bracket without starting bracket");


        for (int i = 0; i < rightSides.size(); i++) {
            Rule errRule = new Rule(null, "0", setOfRightSidePart(rightSides.get(i)), false);
            try {
                ruleService.rulePreProcess(errRule);
                assert false;
            } catch (ResponseStatusException e) {
                assert e.getReason().equals(messages.get(i));
            }
        }

        Rule withAn = new Rule(null, "45", setOfRightSidePart("< xd ><l>"), false);
        ruleService.rulePreProcess(withAn);
        assert withAn.getRightSide().equals(getFromString("< xd ><l>"));

        Rule back = new Rule(null, "3", setOfRightSidePart(" \" xd \" \\\" "), false);
        ruleService.rulePreProcess(back);
        assert back.getRightSide().equals(getFromString("\" xd \" \\\""));

        Rule backNull = new Rule(null, "4", setOfRightSidePart(" \" \\\\\"  "), false);
        ruleService.rulePreProcess(backNull);
        assert backNull.getRightSide().equals(getFromString("\" \\\\\""));

        Rule backAngleBrNull = new Rule(null, "5", setOfRightSidePart(" < \\\\>  "), false);
        ruleService.rulePreProcess(backAngleBrNull);
        assert backAngleBrNull.getRightSide().equals(getFromString("< \\\\>"));

        Rule backAngleBr = new Rule(null, "6", setOfRightSidePart(" < \\> >  "), false);
        ruleService.rulePreProcess(backAngleBr);
        assert backAngleBr.getRightSide().equals(getFromString("< \\> >"));

    }


    @Test
    public void testCommonLeftMerge() {
        Set<Rule> mergeRule = new HashSet<>();
        Rule rule1 = new Rule(null, "xd", setOfRightSidePart("lul"), true);
        Rule rule2 = new Rule(null, "xd", setOfRightSidePart("lul"), true);
        Rule rule3 = new Rule(null, "xd", setOfRightSidePart("ul"), true);
        mergeRule.add(rule1);
        mergeRule.add(rule2);
        mergeRule.add(rule3);

        Rule rule1Merged = ruleService.commonLeftMerge(mergeRule).stream().toList().get(0);
        assert rule1Merged.getRoot();
        assert rule1Merged.getRightSide().equals(getFromString("lul|ul"));
    }

    @Test
    public void testStraightBrackets() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "0", setOfRightSidePart("[xd]"), false);
        ruleService.rulePreProcess(rule);
        assert rule.getRightSide().equals(getFromString("xd|<NULL>"));
        Rule rule1 = new Rule(null, "1", setOfRightSidePart("[a|b]"), false);
        ruleService.rulePreProcess(rule1);
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        for (RightSidePart partString : rule1.getRightSide()) {
            switch (partString.getPart()) {
                case "a" -> one++;
                case "b" -> two++;
                case "<NULL>" -> three++;
                default -> four++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 0;
        one = 0;
        two = 0;
        three = 0;

        Rule rule2 = new Rule(null, "2", setOfRightSidePart("[(a|b) c]"), false);
        rule2 = ruleService.processAddRule(rule2, grammar).getValue0();
        for (RightSidePart partString : rule2.getRightSide()) {
            switch (partString.getPart()) {
                case "a c" -> one++;
                case "b c" -> two++;
                case "<NULL>" -> three++;
                default -> four++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 0;
    }

    @Test
    public void testListDeleteAll() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "1", setOfRightSidePart("xd"), false);
        Rule rule1 = new Rule(null, "2", setOfRightSidePart("d"), true);
        ruleService.processAddRule(rule, grammar);
        ruleService.processAddRule(rule1, grammar);
        int one = 0;
        int two = 0;
        int three = 0;
        List<Rule> rules = ruleService.getAllRules();
        assert rules.size() == 2;
        for (Rule partRule : rules) {
            switch (partRule.getLeftSymbol()) {
                case "<1>" -> one++;
                case "<2>" -> two++;
                default -> three++;
            }
        }
        assert one == 1 && two == 1 && three == 0;
        ruleService.deleteRules();
        assert ruleService.getAllRules().size() == 0;
    }


    @Test
    @Transactional
    public void testUpdateRule() {
        grammarService.deleteAll();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "old", setOfRightSidePart("xd"), false);
        Rule newRule = new Rule(null, "new", setOfRightSidePart("d|d|d"), false);
        Rule root = new Rule(null, "root", setOfRightSidePart("l|S|l"), true);

        rule = ruleService.processAddRule(rule, grammar).getValue0();
        root = ruleService.processAddRule(root, grammar).getValue0();
        Long id = rule.getId();
        newRule = ruleService.updateRule(rule.getId(), newRule);
        assert Objects.equals(id, newRule.getId());
        assert newRule.getRightSide().equals(getFromString("d"));
        assert newRule.getLeftSymbol().equals("<new>");
        root = ruleService.changeRootStatus(root.getId());
        assert !root.getRoot();
        root = ruleService.changeRootStatus(root.getId());
        assert root.getRoot();
        Rule tryNull = new Rule(null, "noch", setOfRightSidePart("noch"), false);
        tryNull = ruleService.processAddRule(tryNull, grammar).getValue0();
        Rule empty = new Rule(null, null, null, null);
        tryNull = ruleService.updateRule(tryNull.getId(), empty);
        assert tryNull.getLeftSymbol().equals("<noch>");
        assert tryNull.getRightSide().equals(getFromString("noch"));
        Rule zeroLen = new Rule(null, "", new HashSet<>(), null);
        tryNull = ruleService.updateRule(tryNull.getId(), zeroLen);
        assert tryNull.getLeftSymbol().equals("<noch>");
        assert tryNull.getRightSide().equals(getFromString("noch"));
        Rule xd = new Rule(null, "xd", setOfRightSidePart("sm"), null);
        xd = ruleService.processAddRule(xd, grammar).getValue0();
        xd = ruleService.updateRule(xd.getId(), new Rule(null, null, setOfRightSidePart("1(1|2)"), null));
        Set<RightSidePart> result = xd.getRightSide();
        for (RightSidePart part : result) {
            assert !part.getPart().contains("(");
        }
        Rule old2 = new Rule(null, "da", setOfRightSidePart("xd"), null);
        Rule old = new Rule(null, "old4", setOfRightSidePart("sm"), null);
        Rule new2 = new Rule(null, "old4", new HashSet<>(), null);
        old = ruleService.processAddRule(old, grammar).getValue0();
        old2 = ruleService.processAddRule(old2, grammar).getValue0();
        old2 = ruleService.updateRule(old2.getId(), new2);
        assert old2.getRightSide().size() == 2;

        Rule old4 = new Rule(null, "dfa", setOfRightSidePart("xd"), null);
        Rule old3 = new Rule(null, "old2", setOfRightSidePart("sm"), null);
        Rule new4 = new Rule(null, "old2", setOfRightSidePart("li"), null);
        old3 = ruleService.processAddRule(old3, grammar).getValue0();
        old4 = ruleService.processAddRule(old4, grammar).getValue0();
        old4 = ruleService.updateRule(old4.getId(), new4);
        assert old4.getRightSide().size() == 2;
        ruleService.getAllRules();
    }

    @Test
    public void testFindDeleteById() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "1", setOfRightSidePart("xd"), false);
        Rule rule1 = new Rule(null, "2", setOfRightSidePart("d"), true);
        Rule rule3 = new Rule(null, "3", setOfRightSidePart("d"), false);
        rule = ruleService.processAddRule(rule, grammar).getValue0();
        rule1 = ruleService.processAddRule(rule1, grammar).getValue0();
        ruleService.processAddRule(rule3, grammar);
        assert ruleService.getById(rule.getId()).getLeftSymbol().equals("<1>");
        ruleService.deleteById(rule1.getId());
        List<Rule> rules = ruleService.getAllRules();
        assert rules.size() == 2;
        int one = 0;
        int def = 0;
        int three = 0;
        for (Rule smallRule : rules) {
            switch (smallRule.getLeftSymbol()) {
                case "<1>" -> one++;
                case "<3>" -> three++;
                default -> def++;
            }
        }
        assert one == 1 && three == 1 && def == 0;


    }

    @Test
    @Transactional
    public void testGetByRightSide() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Grammar grammar2 = new Grammar();
        grammar2 = grammarService.addGrammar(grammar2);
        Rule rule = new Rule(null, "1", setOfRightSidePart("xd"), false);
        Rule rule1 = new Rule(null, "2", setOfRightSidePart("d"), true);
        Rule rule3 = new Rule(null, "1", setOfRightSidePart("d"), false);
        ruleService.processAddRule(rule, grammar);
        ruleService.processAddRule(rule1, grammar);
        ruleService.processAddRule(rule3, grammar2);
        List<Rule> rules = ruleService.getByLeftSide("1");
        assert rules.size() == 2;
        for (Rule smallRule : rules) {
            assert smallRule.getLeftSymbol().equals("<1>");
        }
        List<Rule> rulesFromGrammar = ruleService.getByLeftSide("1", grammar.getId());
        assert rulesFromGrammar.size() == 1;
        for (Rule smallRule : rulesFromGrammar) {
            assert smallRule.getLeftSymbol().equals("<1>") && smallRule.getRightSide().equals(getFromString("xd"));
        }
        assert ruleService.getByLeftSide("5").size() == 0;
    }
}
